from scenarios import *

# At the start of every scenario, we automatically insert a
# sequences.handshake() call. So, this scenario is just here to test that
# this basic thing works fine.

scenario = (
)
